"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const concat_1 = require("./iterators/concat");
const distinct_1 = require("./iterators/distinct");
const exclude_1 = require("./iterators/exclude");
const fill_1 = require("./iterators/fill");
const filter_1 = require("./iterators/filter");
const flat_1 = require("./iterators/flat");
const groupJoin_1 = require("./iterators/groupJoin");
const intersect_1 = require("./iterators/intersect");
const join_1 = require("./iterators/join");
const leftJoin_1 = require("./iterators/leftJoin");
const map_1 = require("./iterators/map");
const reverse_1 = require("./iterators/reverse");
const shuffle_1 = require("./iterators/shuffle");
const skip_1 = require("./iterators/skip");
const slice_1 = require("./iterators/slice");
const splice_1 = require("./iterators/splice");
const take_1 = require("./iterators/take");
const nth_1 = require("./reducers/nth");
const average_1 = require("./reducers/average");
const length_1 = require("./reducers/length");
const first_1 = require("./reducers/first");
const indexOf_1 = require("./reducers/indexOf");
const last_1 = require("./reducers/last");
const lastIndexOf_1 = require("./reducers/lastIndexOf");
const max_1 = require("./reducers/max");
const min_1 = require("./reducers/min");
const reduce_1 = require("./reducers/reduce");
const sum_1 = require("./reducers/sum");
const toArray_1 = require("./reducers/toArray");
const toGroups_1 = require("./reducers/toGroups");
const toMap_1 = require("./reducers/toMap");
const toSet_1 = require("./reducers/toSet");
const isIterable_1 = require("./utils/isIterable");
const iterable_1 = require("./utils/iterable");
const iterator_1 = require("./utils/iterator");
/**
 * Creates a queryable iterable.
 * @param source can be an array or any other iterable.
 */
function query(source) {
    return new Query(source);
}
exports.query = query;
class Query {
    constructor(source) {
        this.source = source;
    }
    [Symbol.iterator]() {
        return iterator_1.iterator(this.source);
    }
    // #region common methods
    entries() {
        return new Query(map_1.map(this, (elem, idx) => [idx, elem]));
    }
    keys() {
        return new Query(map_1.map(this, (elem, idx) => idx));
    }
    values() {
        return new Query(this);
    }
    forEach(action) {
        for (const [index, value] of this.entries()) {
            action(value, index);
        }
    }
    concat(other) {
        return isIterable_1.isIterable(other) ?
            new Query(concat_1.concat(this, other)) :
            new Query(concat_1.concat(this, [other]));
    }
    prepend(other) {
        return isIterable_1.isIterable(other) ?
            new Query(concat_1.concat(other, this)) :
            new Query(concat_1.concat([other], this));
    }
    fill(value, start, end) {
        return new Query(fill_1.fill(this, value, start, end));
    }
    // #endregion
    // #region IterableValue implementation
    nth(index) {
        return nth_1.nth(this, index);
    }
    indexOf(element, fromIndex = 0) {
        return indexOf_1.indexOf(this, (elem, idx) => idx >= fromIndex && elem === element);
    }
    findIndex(predicate) {
        return indexOf_1.indexOf(this, predicate);
    }
    lastIndexOf(element, fromIndex = 0) {
        return lastIndexOf_1.lastIndexOf(this, (elem, idx) => idx >= fromIndex && elem === element);
    }
    findLastIndex(predicate) {
        return lastIndexOf_1.lastIndexOf(this, predicate);
    }
    length(predicate = alwaysTrue()) {
        return length_1.length(filter_1.filter(this, predicate));
    }
    first() {
        return first_1.first(this);
    }
    find(predicate) {
        return first_1.first(filter_1.filter(this, predicate));
    }
    last() {
        return last_1.last(this);
    }
    findLast(predicate) {
        return last_1.last(filter_1.filter(this, predicate));
    }
    average(selector = element()) {
        return average_1.average(map_1.map(this, selector));
    }
    min(compareFn = comparer()) {
        return min_1.min(this, compareFn);
    }
    max(compareFn = comparer()) {
        return max_1.max(this, compareFn);
    }
    sum(selector = element()) {
        return sum_1.sum(map_1.map(this, selector));
    }
    reduce(callback, initialValue) {
        return reduce_1.reduce(this, callback, initialValue);
    }
    reduceRight(callback, initialValue) {
        return reduce_1.reduce(reverse_1.reverse(this), callback, initialValue);
    }
    // #endregion
    // #region IterablePredicate implementation
    every(predicate) {
        return indexOf_1.indexOf(this, (e, i) => !predicate(e, i)) === -1;
    }
    some(predicate) {
        return indexOf_1.indexOf(this, predicate) !== -1;
    }
    includes(element, fromIndex = 0) {
        return this.some((elem, idx) => idx >= fromIndex && elem === element);
    }
    // #endregion
    // #region IterablePermutation implementation
    sort(compareFn = comparer()) {
        return new Query((function* (source) { yield* source.toArray().sort(compareFn); })(this));
    }
    shuffle() {
        return new Query(shuffle_1.shuffle(this));
    }
    reverse() {
        return new Query(reverse_1.reverse(this));
    }
    // #endregion
    // #region IterableFilter implementation
    filter(predicate) {
        return new Query(filter_1.filter(this, predicate));
    }
    take(count) {
        return new Query(take_1.take(this, count));
    }
    skip(count) {
        return new Query(skip_1.skip(this, count));
    }
    slice(start, end) {
        return new Query(slice_1.slice(this, start, end));
    }
    splice(start, deleteCount, ...items) {
        return new Query(splice_1.splice(this, start, deleteCount, items));
    }
    // #endregion
    // #region IterableTransformation implementation
    map(selector) {
        return new Query(map_1.map(this, selector));
    }
    flat(selector) {
        return new Query(flat_1.flat(this.map(selector)));
    }
    groupBy(keySelector, valueSelector = x => x) {
        const groups = iterable_1.iterable(() => toGroups_1.toGroups(this, keySelector, valueSelector));
        const result = map_1.map(groups, elem => [elem[0], new Query(elem[1])]);
        return new Query(result);
    }
    // #endregion
    // #region IterableSet implementation
    distinct(selector = element()) {
        return new Query(distinct_1.distinct(this, selector));
    }
    exclude(other, selector = element()) {
        return new Query(exclude_1.exclude(this, other, selector));
    }
    intersect(other, selector = element()) {
        return new Query(intersect_1.intersect(this, other, selector));
    }
    union(other, selector = element()) {
        return new Query(distinct_1.distinct(concat_1.concat(this, other), selector));
    }
    // #endregion
    // #region IterableJoin implementation
    join(other, leftKeySelector, rightKeySelector, joinSelector) {
        const iterator = join_1.join(this, other, leftKeySelector, rightKeySelector, joinSelector);
        return new Query(iterator);
    }
    leftJoin(other, leftKeySelector, rightKeySelector, joinSelector) {
        const iterator = leftJoin_1.leftJoin(this, other, leftKeySelector, rightKeySelector, joinSelector);
        return new Query(iterator);
    }
    rightJoin(other, rightKeySelector, leftKeySelector, joinSelector) {
        const iterator = leftJoin_1.leftJoin(other, this, rightKeySelector, leftKeySelector, joinSelector);
        return new Query(iterator);
    }
    groupJoin(other, leftKeySelector, rightKeySelector, joinSelector) {
        const iterator = groupJoin_1.groupJoin(this, other, leftKeySelector, rightKeySelector, joinSelector);
        return new Query(iterator);
    }
    // #endregion
    // #region IterableCast implementation
    toArray(selector = element()) {
        return toArray_1.toArray(map_1.map(this, selector));
    }
    toMap(keySelector = element(), valueSelector = element()) {
        return toMap_1.toMap(this, keySelector, valueSelector);
    }
    toGroups(keySelector = element(), valueSelector = element()) {
        return toGroups_1.toGroups(this, keySelector, valueSelector);
    }
    toSet(selector = element()) {
        return toSet_1.toSet(map_1.map(this, selector));
    }
    toString() {
        return this.toArray().toString();
    }
}
function element() {
    return (e, index) => e;
}
function alwaysTrue() {
    return (e) => true;
}
function comparer() {
    return (a, b) => a === b ? 0 : (a > b ? 1 : -1);
}

//# sourceMappingURL=Query.js.map
